﻿using Mono.Cecil.Cil;
using MonoMod.Cil;
using System;
using System.Reflection;
using Terraria;
using Terraria.ModLoader;

namespace TigerTester.Systems;

public class ZeroDamageOnNPCSystem : ModSystem {
    public enum BoolWithFullyDisable {
        FullyDisable,
        Disable,
        Enable
    }
    // 实际上应该获取本模组实例
    // 这里主要用作报错
    public static Mod ModInstance => null;

    #region 上下钩子
    public override void OnModLoad() {
        try {
            MonoModHooks.Modify(npcHitInfoSetDamage, IL_NPC_HitInfo_set_Damage);
            MonoModHooks.Modify(npcHitInfoSetSourceDamage, IL_NPC_HitInfo_set_SourceDamage);
            IL_NPC.StrikeNPC_HitInfo_bool_bool += IL_NPC_StrikeNPC;
            IL_NPC.HitModifiers.ToHitInfo += IL_NPC_HitModifiers_ToHitInfo;
            IL_NPC.HitModifiers.GetDamage += IL_NPC_HitModifiers_GetDamage;

            if(ProjectileHook != BoolWithFullyDisable.FullyDisable) {
                HookProjectile();
            }
        }
        catch(Exception e) {
            ModInstance.Logger.Error("Add IL Error!", e);
        }
    }
    public override void OnModUnload() {
        try {
            // 实际上不需要手动卸载
            // UnhookProjectile();
        }
        catch(Exception e) {
            ModInstance.Logger.Error("Remove IL Error!", e);
        }
    }
    #endregion
    
    #region 缓存一些反射内容
    const BindingFlags bfps = BindingFlags.Public | BindingFlags.Static;
    const BindingFlags bfpi = BindingFlags.Public | BindingFlags.Instance;
    const BindingFlags bfns = BindingFlags.NonPublic | BindingFlags.Static;
    const BindingFlags bfni = BindingFlags.NonPublic | BindingFlags.Instance;
    static readonly FieldInfo npcHitModifiersSourceDamage = typeof(NPC.HitModifiers).GetField(nameof(NPC.HitModifiers.SourceDamage))!;
    static readonly MethodInfo statModifierApplyTo = typeof(StatModifier).GetMethod(nameof(StatModifier.ApplyTo))!;
    static readonly MethodInfo mathMaxInt = typeof(Math).GetMethod(nameof(Math.Max), bfps, [typeof(int), typeof(int)])!;
    static readonly MethodInfo npcHitInfoSetSourceDamage = typeof(NPC.HitInfo).GetProperty(nameof(NPC.HitInfo.SourceDamage))!.GetSetMethod()!;
    static readonly FieldInfo npcHitModifiers_damageLimit= typeof(NPC.HitModifiers).GetField("_damageLimit", bfni)!;
    static readonly MethodInfo mathMinInt = typeof(Math).GetMethod(nameof(Math.Min), bfps, [typeof(int), typeof(int)])!;
    static readonly MethodInfo mathClampInt = typeof(Math).GetMethod(nameof(Math.Clamp), bfps, [typeof(int), typeof(int), typeof(int)])!;
    static readonly MethodInfo mathMaxFloat = typeof(Math).GetMethod(nameof(Math.Max), bfps, [typeof(float), typeof(float)])!;
    static readonly MethodInfo npcHitInfoSetDamage = typeof(NPC.HitInfo).GetProperty(nameof(NPC.HitInfo.Damage))!.GetSetMethod()!;
    #endregion

    #region NPC.StrikeNPC
    /// <summary>
    /// <c>if (damage >= 1.0)</c> 改为 <c>if (damage >= 0.0)</c>
    /// </summary>
    static void IL_NPC_StrikeNPC(ILContext il) {
        try {
            ILCursor cursor = new(il);
            ILLabel? bleLabel = null;
            cursor.GotoNext(MoveType.AfterLabel,
                i => i.MatchLdloc1(),
                i => i.MatchLdcR8(1.0),
                i => i.MatchBltUn(out bleLabel));
            cursor.EmitLdloc1();
            cursor.EmitLdcR8(0.0);
            // il.Instrs[cursor.Index + 2].MatchBltUn(out var bleLabel);
            // 使用 Target 而不是直接使用标签作为参数是为了创建一个新的标签出来, 防止耦合
            cursor.EmitBltUn(bleLabel!.Target!);
            cursor.EmitBr(il.Instrs[cursor.Index + 3]);
        }
        catch(Exception e) {
            ModInstance.Logger.Error("IL NPC.StrikeNPC Error!", e);
        }
    }
    #endregion

    #region NPC.HitModifiers.ToHitInfo
    /// <summary>
    /// <see cref="NPC.HitModifiers.ToHitInfo"/>时限制 SourceDamage 的最小值为 0 而不是 1
    /// </summary>
    static void IL_NPC_HitModifiers_ToHitInfo(ILContext il) {
        try {
            ILCursor cursor = new(il);
            cursor.GotoNext(MoveType.AfterLabel,
                i => i.MatchLdloca(1),
                i => i.MatchLdarg0(),
                i => i.MatchLdflda(npcHitModifiersSourceDamage),
                i => i.MatchLdarg1(),
                i => i.MatchCall(statModifierApplyTo),
                i => i.MatchConvI4(),
                i => i.MatchLdcI4(1),
                i => i.MatchCall(mathMaxInt),
                i => i.MatchCall(npcHitInfoSetSourceDamage));
            for(Instruction i = cursor.Next!; !i.MatchLdcI4(1); i = i.Next) {
                cursor.Emit(i.OpCode, i.Operand);
            }
            cursor.EmitLdcI4(0);
            cursor.EmitCall(mathMaxInt);
            cursor.EmitCall(npcHitInfoSetSourceDamage);
            cursor.EmitBr(il.Instrs[cursor.Index + 9]);
        }
        catch(Exception e) {
            ModInstance.Logger.Error("IL NPC.HitModifiers.ToHitInfo Error!", e);
        }
    }
    #endregion

    #region NPC.HitModifiers.GetDamage
    static void IL_NPC_HitModifiers_GetDamage(ILContext il) {
        try {
            #region 第 1 部分
            ILCursor cursor = new(il);
            cursor.GotoNext(MoveType.AfterLabel,
                i => i.MatchLdloc(5),
                i => i.MatchConvI4(),
                i => i.MatchLdcI4(1),
                i => i.MatchLdarg0(),
                i => i.MatchLdfld(npcHitModifiers_damageLimit),
                i => i.MatchLdcI4(4),
                i => i.MatchCall(mathMinInt),
                i => i.MatchCall(mathClampInt),
                i => i.MatchRet());
            for(Instruction i = cursor.Next!; !i.MatchRet(); i = i.Next) {
                if(i.MatchLdcI4(1)) {
                    cursor.EmitLdcI4(0);
                    continue;
                }
                cursor.Emit(i.OpCode, i.Operand);
            }
            cursor.EmitBr(il.Instrs[cursor.Index + 8]);
            #endregion

            #region 第 2 部分
            cursor.GotoNext(MoveType.AfterLabel,
                i => i.MatchLdloc0(),
                i => i.MatchLdloc3(),
                i => i.MatchSub(),
                i => i.MatchLdcR4(1),
                i => i.MatchCall(mathMaxFloat),
                i => i.MatchStloc0());
            cursor.EmitLdloc0();
            cursor.EmitLdloc3();
            cursor.EmitSub();
            cursor.EmitLdcR4(0);
            cursor.EmitCall(mathMaxFloat);
            cursor.EmitStloc0();
            cursor.EmitBr(il.Instrs[cursor.Index + 6]);
            #endregion

            #region 第 3 部分
            cursor.GotoNext(MoveType.AfterLabel,
                i => i.MatchLdloca(6),
                i => i.MatchLdloc0(),
                i => i.MatchCall(statModifierApplyTo),
                i => i.MatchConvI4(),
                i => i.MatchLdcI4(1),
                i => i.MatchLdarg0(),
                i => i.MatchLdfld(npcHitModifiers_damageLimit),
                i => i.MatchCall(mathClampInt),
                i => i.MatchRet());
            for(Instruction i = cursor.Next!; !i.MatchRet(); i = i.Next) {
                if(i.MatchLdcI4(1)) {
                    cursor.EmitLdcI4(0);
                    continue;
                }
                cursor.Emit(i.OpCode, i.Operand);
            }
            cursor.EmitBr(il.Instrs[cursor.Index + 8]);
            #endregion
        }
        catch(Exception e) {
            ModInstance.Logger.Error("IL NPC.HitModifiers.GetDamage Error!", e);
        }
    }
    #endregion

    #region NPC.HitInfo.set_[Source]Damage
    static void IL_NPC_HitInfo_set_Damage(ILContext il) {
        try {
            ILCursor cursor = new(il);
            cursor.GotoNext(MoveType.After, i => i.MatchLdcI4(1));
            cursor.EmitPop();
            cursor.EmitLdcI4(0);
        }
        catch(Exception e) {
            ModInstance.Logger.Error("IL NPC.HitInfo.set_Damage Error!", e);
        }
    }
    static void IL_NPC_HitInfo_set_SourceDamage(ILContext il) {
        try {
            ILCursor cursor = new(il);
            cursor.GotoNext(MoveType.After, i => i.MatchLdcI4(1));
            cursor.EmitPop();
            cursor.EmitLdcI4(0);
        }
        catch(Exception e) {
            ModInstance.Logger.Error("IL NPC.HitInfo.set_SourceDamage Error!", e);
        }
    }
    #endregion

    #region Projectile.Damage
    // 实际上应该从配置中获取
    public static BoolWithFullyDisable ProjectileHook => BoolWithFullyDisable.Enable;

    public static bool ZeroProjectileCanDamage => ProjectileHook == BoolWithFullyDisable.Enable;

    // 在配置的 OnChange() 中调用以下两个方法
    // 在将 ZeroDamageProjectileCanDamage 设置为
    // FullyDisable 时调用 Unhook, 否则调用 Hook
    static bool projectileHooked;
    public static void HookProjectile() {
        if(projectileHooked) {
            return;
        }
        projectileHooked = true;
        try {
            IL_Projectile.Damage += MakeZeroDamageProjectileCanDamage;
        }
        catch(Exception e) {
            ModInstance.Logger.Error("Add IL to Projectile.Damage Error!", e);
        }
    }
    public static void UnhookProjectile() {
        if(!projectileHooked) {
            return;
        }
        projectileHooked = false;
        try {
            IL_Projectile.Damage -= MakeZeroDamageProjectileCanDamage;
        }
        catch(Exception e) {
            ModInstance.Logger.Error("Remove IL to Projectile.Damage Error!", e);
        }
    }

    /// <summary>
    /// 让伤害为0的弹幕也能造成伤害
    /// </summary>
    static void MakeZeroDamageProjectileCanDamage(ILContext il) {
        try {
            ILCursor cursor = new(il);
            cursor.GotoNext(MoveType.AfterLabel, i => i.MatchLdloc(10), il => il.MatchBrfalse(out _));
            cursor.EmitLdarg0();
            cursor.EmitLdloc(10);
            cursor.EmitDelegate((Projectile p, bool flag) => ZeroProjectileCanDamage ? p.damage >= 0 : flag);
            cursor.EmitBr(il.Instrs[cursor.Index + 1]);
        }
        catch(Exception e) {
            ModInstance.Logger.Error("IL MakeZeroDamageProjectileCanDamage Error!", e);
        }
    }
    #endregion
}
